/*****************************************************************
*
* PROJECT:        Swiss Pairing Test Program
*
*                 Copyright (C) 1996 Robert Rozycki
*                 SwissPerfect/SPAR General Public Licence
*
* FILE:           spar.cpp
*
* DESCRIPTION:    Main module.
*
* AUTHOR:         Robert Rozycki   rozycki@perth.dialix.oz.au
*
* PORTABILITY:    ANSI C++
*
* $Revision: 1.2 $
*
* $Date: 1996/02/04 09:45:16 $
*
* $Log: spar.cpp $
// Revision 1.2  1996/02/04  09:45:16  robert
// Internet release Feb 1996
//
// Revision 1.1  1996/01/01  14:03:15  robert
// Initial revision
//
*
*******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "l_types.h"
#include "log.h"
#include "spar.h"
#include "utils.h"
#include "sws_pair.h"
#include "sws_read.h"


// Global variables
SwissRead_c *swiss;        // instance of the SwissRead_c class
DebugLog_c   *MsgMngr;


// Local function prototypes
static void PrintFinalPairing( PAIR *pairing, int players );
static void PrintSwiss( char *test_file );
static void PrintUsage();


//////////////////////////////////////////////////////////////////////
//
//  main() function
//  Command line parameters are file names containing Swiss test data
//  or switches 
//
int main( int argc, char *argv[] )
{
   FILE *fdata;
   PAIR *pairing;
   int i, result, pairs_count, players_count, test_files_count;
   char **test_files;
   char *log_file = NULL;
   int pairing_mode = 0;  // normal pairing
   int log_debug_lev = LOG_MED, disp_debug_lev = LOG_MED;

   if( argc < 2 )
   {
      PrintUsage();
      return( 0 );
   }

   test_files_count = 0;
   test_files = (char **)malloc( argc * sizeof(char *) );
 // read filenames & switches
   for( i = 1; i < argc; i++ )
   {
      if( argv[i][0] != '-' )
      {
         test_files[ test_files_count ] = strdup( argv[i] );
         test_files_count++;
      }
      else
      {
         switch( argv[i][1] )
         {
            case 'g':
               pairing_mode = 1;
               break;
            case 'd':
               if( !isdigit( argv[i+1][0] ) )
                  break;
               log_debug_lev = atoi( argv[i+1] );
               i++;
               if( !isdigit( argv[i+1][0] ) )
                  break;
               disp_debug_lev = atoi( argv[i+1] );
               i++;
               break;
            case 'l':
               if( i < argc-1 && argv[i+1][0] != '-' )
               {
                  log_file = strdup( argv[i+1] );
                  i++;
               }
               break;
            default : 
               continue;
         }
      }
   }
   if( log_file == NULL )
       log_file = "spar.log";

   for( i = 0; i < test_files_count; i++ )
   {
      MsgMngr = new DebugLog_c();
      MsgMngr->SetLogFileName( log_file );
      MsgMngr->SetDebugLevs( log_debug_lev, disp_debug_lev );
      MsgMngr->WriteToLog( LOG_FLOW, SPAR_VERSION );
      swiss   = new SwissRead_c( MsgMngr );
     // open input file
      fdata = fopen( test_files[i], "rt" );
     // read in and process Swiss Test data file
      if( fdata == NULL || swiss->ReadData( fdata  ) != SUCCESS )
         exit( EXIT_FAILURE );
      fclose( fdata );

      PrintSwiss( test_files[i] );

      printf( "Running...\n" );
      players_count =  swiss->GetPlayersCount();
      pairs_count = players_count >> 1 ;
      pairing = (PAIR *)malloc( sizeof( PAIR ) * (pairs_count+2) );
      memset( pairing, 0, sizeof(PAIR) * (pairs_count+2) );
   // third parameter must be > 2 to avoid special colour
   // assignment for round 1 and make sure that floats are recognised
   // correctly.
      result = swiss->MakePairing( pairing, players_count, 3, pairing_mode );
      if( result != SUCCESS )
         result = -1;
      else
         result = players_count;
      if( result == -1 )
      {
         MsgMngr->WriteToLog( LOG_FLOW, "Pairing failed." );
         exit(0);
      }
      PrintFinalPairing( pairing, players_count );
      free( pairing );
      delete( swiss );
   }
   return(0);
}


//////////////////////////////////////////////////////////////////////
//
//   Prints processed Swiss data to the log file
//
void PrintSwiss(  char *filename )
{
   int i, j;
   char buf[ 255 ];
   char *p;

   MsgMngr->WriteToLog( LOG_LOW, "Test file name : %s", filename );
 /* print all players their opponents, colour preferences, float status */
   p = buf;
   for( i = 0; i < swiss->GetPlayersCount(); i++ )
   {
      sprintf( buf, "%d:", i+1 );
      for( j = 0; j < swiss->GetOpponCount( i+1 ); j++ )
      {
         p = buf + strlen( buf );
         sprintf( p, " %d,", swiss->GetOppon( i+1, j ) );
      }
      p = buf + strlen( buf );
      sprintf( p, ": Col Pref.%d", swiss->GetColPref( i+1 ) );
      p = buf + strlen( buf );
      sprintf( p, ": DownFl. %d", swiss->GetDnfStatus( i+1 ) );
      p = buf + strlen( buf );
      sprintf( p, ": UpFl. %d", swiss->GetUpfStatus( i+1 ) );
      MsgMngr->WriteToLog( LOG_LOW, buf );
  }
}

//////////////////////////////////////////////////////////////////////
//
//   Prints resulting pairing to the screen and the log file
//
static void  PrintFinalPairing( PAIR *pairing, int players )
{
   int i, j, no, pairs, unpaired = 0;
   int found;
   int left_player, right_player;
   char buf[255];

   buf[0] = 0;
   if( players < 0 )
   {
      MsgMngr->WriteToLog( LOG_FLOW, "Pairing Failed" );
      return;
   }
   else
      MsgMngr->WriteToLog( LOG_FLOW, "Final Pairing" );

   pairs = ( players >> 1 ) + ( players & 0x01 );
   for( i = 1; i <= pairs; i++ )
   {
      left_player  = pairing[i].left;
      right_player = pairing[i].right;
      if( left_player == 0 )
      {
         pairs =  i - 1 ;
         unpaired = players - ( pairs << 1 );
         break;
      }
      if( right_player != 0 )
         sprintf( buf+strlen(buf), "%3d-%d ", left_player, right_player );
      else
         sprintf( buf+strlen(buf), " float %d ", left_player );
      if( strlen( buf ) > 60 )
      {
         MsgMngr->WriteToLog( LOG_HIGH, buf );
         buf[0] = 0;
      }
   }
   if( strlen( buf ) )
   {
       MsgMngr->WriteToLog( LOG_HIGH, buf );
       buf[0] = 0;
   }
   if( unpaired )
   {
      for( i = 1; i <= players; i++ )
      {
         no = swiss->DemapPlayer( i );
         found = FALSE;
         for( j = 1; j <= pairs; j++ )
         {
            if( pairing[j].left == no || pairing[j].right == no  )
            {
               found = TRUE;
               break;
            }
         }
         if( found == FALSE )
         {
            sprintf( buf+strlen(buf), "%3d ", swiss->DemapPlayer( i ) );
            if( strlen( buf ) > 60 )
            {
               MsgMngr->WriteToLog( LOG_HIGH, buf );
               buf[0] = 0;
            }
         }
      }
   }
   if( strlen( buf ) )
   {
      MsgMngr->WriteToLog( LOG_HIGH, buf );
      buf[0] = 0;
   }

   MsgMngr->WriteToLog( LOG_HIGH, "\nPairing Summary: pairs = %d, unpaired = %d\n",
                           pairs, unpaired );
   MsgMngr->WriteToLog( LOG_HIGH, "\n" );
}

//////////////////////////////////////////////////////////////////////
//
//   Prints usage information 
//
static void PrintUsage()
{
   printf( SPAR_VERSION );
   printf("\nUsage: spar filename [filename] ... [filename]");
   printf("\nOptions:");
   printf("\n   -l [filename]  log file name" );
   printf("\n   -g             single group pairing mode" );
   printf("\n   -d n1 n2       debug levels ( n1 - log_file, n2 - screen )");
   printf("\n\n");
}

