/*****************************************************************
*
* PROJECT:     Swiss Perfect 3
*
*              Copyright (C) 1996 Robert Rozycki
*              SwissPerfect/SPAR General Public Licence
*
* FILE:        sws_ctrl.cpp 
*
* DESCRIPTION: This module controls the swiss pairing process
*              It calls functions from sws_pair.cpp (group pairing)
*
* AUTHOR:      Robert Rozycki   rozycki@perth.dialix.oz.au 
*
* PORTABILITY: ANSI C++
* 
* $Revision: 1.1 $
*
* $Date: 1996/02/04 09:52:03 $
*
* $Log: sws_ctrl.cpp $
// Revision 1.1  1996/02/04  09:52:03  robert
// Initial revision
//
*
*******************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "sws_ctrl.h"


/*========================  I N T E R F A C E ===========================*/

/*-----------------------------------------------------------------------*/
/*
** FUNCTION:     MakePairing()
**
** DESCRIPTION:  Controls the entire process of pairing (group by group)
**
** PARAMETERS:   pairing_buf   - buffer to receive the resulting pairing 
**               players_count - number of players to be paired
**               group_only    - TRUE if we are pairing just one group
**                               (used by test programs only) and
**                               FALSE otherwise (default)
**
** RETURNS:      SUCCESS    - if pairing succeeded
**               FAILURE    - if pairing failed
**               OUT_OF_MEM - on out of memory condition
*/
/*-----------------------------------------------------------------------*/
int SwissCtrl_c::MakePairing( PAIR *pairing_buf, int players_count,
                              int round, int group_only )
{
   int j, group_size, rx_floats_count;
   int gr_pair_count, curr_group, result;
   int *pRxFloats;
   int err_no = 0;
   int last_group = FALSE;
   int from_pair;

   PlayersToBePaired  = players_count;
   PairTablesToBeMade = (players_count >> 1) + (players_count & 0x01);
   RoundNoToPair      = round;
   if( InitPairing( ) != SUCCESS )
      return( OUT_OF_MEM );
   MsgMngr->WriteToLog( LOG_MED,
             "MakePairing(): players=%d, groups=%d, tables=%d, round=%d",
              PlayersToBePaired, GroupsCount, PairTablesToBeMade,
              RoundNoToPair );
   if( PlayersToBePaired & 0x01 )
      AssignBye();
   PairedGrMarks[ 0 ] = 0;
   rx_floats_count = 0;
   pRxFloats = NULL;
   curr_group = 1;   //### if GroupsCount == 0 crashes

// main loop:  pairs group after group
   while( curr_group <= GroupsCount && last_group == FALSE )
   {
      MsgMngr->WriteToLog( LOG_MED,"Pairing group no %d", curr_group );
      if( curr_group == GroupsCount )
         last_group = TRUE;
      group_size = BuildGroupVector( curr_group, rx_floats_count, pRxFloats );
      if( group_size <= 1 )
      {
         gr_pair_count = 0;
         result = 0;
      }
      else
      {
         if( group_only )  // for testing of one group pairings
            result = PairGroup( &GroupPtr, FALSE );
         else
            result = PairGroup( &GroupPtr, last_group );
         gr_pair_count = GroupPtr.count;
         if( last_group == TRUE && result != group_size/2 && !group_only)
            result = -1;
      }
      if( result < 0 )
      {
         err_no = gr_pair_count;
         err_no = ComboPairing( curr_group-1, &group_size );
         if( err_no == -1 )
            return( err_no );
         else
            last_group = TRUE;
// ### shouldn't it be last_round = LAST_GROUP here
         gr_pair_count = GroupPtr.count;
         // ### Temp Fix for Final Pairs
         PairedGrMarks[curr_group-1] = (PlayersToBePaired >> 1) - gr_pair_count;
      }
      else
      {
         for( j = 1; j <= 2 * gr_pair_count; j++ )
            AlreadyPaired[ GroupPtr.group[ j ] ] = TRUE;
      }

      from_pair = PairedGrMarks[curr_group-1] + 1;
      LoadFinalPairs( GroupPtr.group, from_pair, gr_pair_count );
      LogFinalPairs( from_pair, from_pair + gr_pair_count - 1 );
      rx_floats_count = group_size - ( gr_pair_count << 1 );
      PairedGrMarks[curr_group] = PairedGrMarks[curr_group-1]+gr_pair_count;
      free( pRxFloats );
      pRxFloats = NULL;
      if( rx_floats_count != 0 )
      {
         pRxFloats = (int *)malloc( (rx_floats_count + 1) * sizeof(int) );
         if( pRxFloats == NULL )
            return( OUT_OF_MEM );
         for( j = 1; j <= rx_floats_count; j++ )
            pRxFloats[ j ] = GroupPtr.group[ 2 * gr_pair_count + j ];
      }
      free( GroupPtr.group );
      curr_group++;
   }
   SaveFinalPairing( pairing_buf );
   free( pRxFloats );

   return( SUCCESS );
}

/*=================== I M P L E M E N T A T I O N ======================*/

//////////////////////////////////////////////////////////////////////
//
// This function is called when we have an odd number of players
// to assign a bye to one of them.  
// First it looks for players who have not had byes before (or
// haven't been downfloated in the last two rounds).
// Then relaxes the criteria until someone is found.
// NOTE: This function should never fail.
//
int SwissCtrl_c::AssignBye( )
{
   int player_no, i = 0;
   int last_pair[ 3 ]; 

   while( ( player_no = FindPlayerForBye( i ) ) == -1 )
      i++;
   AlreadyPaired[player_no] = TRUE;
   last_pair[ 1 ] = player_no;
   last_pair[ 2 ] = 0;
   LoadFinalPairs( last_pair, PairTablesToBeMade, 1 );
   MsgMngr->WriteToLog( LOG_MED, "AssignBye(): player=%d", player_no );
   return SUCCESS;
}

//////////////////////////////////////////////////////////////////////
//
//  This function is called by AssignBye() to find a prospective Bye
//  holder.
//
int SwissCtrl_c::FindPlayerForBye( int allowed_byes )
{
   int i, bye_status, dnf_status;
   int found = FALSE, found_no = 0, found_dnf_status = 0;

   for( i = PlayersToBePaired; i >= 1; i-- )
   {
      if( ( bye_status = GetByeStatus(i) ) <= allowed_byes )
      {
          dnf_status = GetDnfStatus(i);
          if( bye_status < allowed_byes || dnf_status == 0 ) // ever happens?
             return i;
          if( found == FALSE || dnf_status < found_dnf_status )
          {
             found = TRUE;
             found_no = i;
             found_dnf_status = dnf_status;
          }
      }
   }
   if( found == TRUE && found_no > 0 )  // second check is in fact redundant
      return found_no;
   else
      return -1;
}

//////////////////////////////////////////////////////////////////////
//
//  Allocates memory for object's arrays
//  Returns OUT_OF_MEM if allocation fails
//
int SwissCtrl_c::AllocTables( void )
{
  FinalPairing = (PAIR *)malloc( (PairTablesToBeMade+1)* sizeof(PAIR) );
  if( FinalPairing == NULL )
     return( OUT_OF_MEM );
  memset( FinalPairing, 0, ( PairTablesToBeMade+1 ) * sizeof(PAIR) );

  AlreadyPaired = (char *)malloc( ( PlayersToBePaired + 1 )* sizeof(char) );
  if( AlreadyPaired == NULL )
     return( OUT_OF_MEM );
  memset( AlreadyPaired, 0, ( PlayersToBePaired+1 ) * sizeof( char ) );

  return( SUCCESS );
}

//////////////////////////////////////////////////////////////////////
//
//  Initialises object's data, retrieves group marks.
//
int SwissCtrl_c::InitPairing( )
{
  int first, last, result;

  GroupsCount = 0;
  if( AllocTables() != SUCCESS )
      return( OUT_OF_MEM );
  for(;;)
  {
     result = GetGroup( GroupsCount, first, last );
     if( result != SUCCESS )
        break;
     GroupMarks[ GroupsCount ] = first;
     GroupsCount++;
  }
  GroupMarks[ GroupsCount ] = PlayersToBePaired + 1;
  return( SUCCESS );
}

//////////////////////////////////////////////////////////////////////
//
//   This function is called after any group pairing is completed.
//   Loads created pairs from vector "group" into FinalPairing[]
//
//   Parameters:   group      - pointer to group list
//                 start_pair - first free pair in FinalPairing[]
//                 pairs_made - number of pairs to be loaded
//
void SwissCtrl_c::LoadFinalPairs( int *group, int start_pair, int pairs_made )
{
  int i, idx;

  idx = 1;
  for( i = start_pair; i < start_pair + pairs_made; i++ )
  {
     FinalPairing[ i ].left  = group[ idx++ ];
     FinalPairing[ i ].right = group[ idx++ ];
  }
}

//////////////////////////////////////////////////////////////////////
//
//  Unpairs groups paired previously
//
//  Parameter:   group_no  - group number at which start unpairing
//
int SwissCtrl_c::UnpairGroup( int group_no )
{
   int i;

   MsgMngr->WriteToLog( LOG_MED, "UnpairGroup(): %d, pairs:", group_no );

   i = PairedGrMarks[ group_no - 1 ] + 1;
   LogFinalPairs( i, PairTablesToBeMade );
   while( i <= PairTablesToBeMade )
   {
      if( FinalPairing[ i ].left == 0 )
          break;
      AlreadyPaired[ FinalPairing[ i ].left ] = FALSE;
      AlreadyPaired[ FinalPairing[ i ].right ] = FALSE;
      FinalPairing[ i ].left = 0;
      FinalPairing[ i ].right = 0;
      i++;
   }
   return( SUCCESS );
}

//////////////////////////////////////////////////////////////////////
//
// This function is called when we have to complete pairing
// by pairing all remaining players as one group ('Combo Group')
//
int SwissCtrl_c::ComboBuildGroupVector( void )
{
   int i, index, curr_group ;
   int group_size = 0;

   curr_group = 1;
   for( i = 1; i <= PlayersToBePaired; i++ )
   {
      if( AlreadyPaired[ i ] == FALSE )
          group_size++;
   }
   GroupPtr.group = (int *)malloc( ( group_size + 1 ) * sizeof( int ) );
   if( GroupPtr.group == NULL )
      return( OUT_OF_MEM );
   GroupPtr.brackets = (char *)malloc( ( group_size+1 ) * sizeof( char ) );
   if( GroupPtr.brackets == NULL )
      return( OUT_OF_MEM );
   index = 0;
   curr_group = 1;
   for( i = 1; i <= PlayersToBePaired; i++ )
   {
      if( AlreadyPaired[ i ] == FALSE )
      {
          index++;
          GroupPtr.group[ index ] = i;
          GroupPtr.brackets[ index ] = curr_group;
      }
      if( i == GroupMarks[ curr_group ] )
         curr_group++;
   }
   GroupPtr.count = group_size;
   return( group_size );
}

//////////////////////////////////////////////////////////////////
//
//   Loads current floaters and current group players into *GroupPtr
//
//  PARAMETERS:   group_no        - group number at which start unpairing
//                rx_floats_count - number of downfloaters in the group
//                pRxFloats       - pointer to an array holding current
//                                  floaters
//
//  RETURNS:      number of players to be paired
//
int SwissCtrl_c::BuildGroupVector( int group_no, int rx_floats_count,
                                   int *pRxFloats )
{
   int j, player, index;
   int group_size = 0;

 // free if not NULL
   GroupPtr.brackets = NULL;
   group_size = GroupMarks[ group_no ] - GroupMarks[ group_no-1 ];
   group_size += rx_floats_count;
   GroupPtr.group = (int *)malloc( ( group_size + 1 ) * sizeof( int ) );
   if( GroupPtr.group == NULL )
       return( OUT_OF_MEM );
   for( j = 1; j <= rx_floats_count; j++ )
       GroupPtr.group[ j ] = pRxFloats[ j ];
   player = GroupMarks[ group_no - 1 ];
   index = rx_floats_count + 1;
   for( j = rx_floats_count + 1; j <= group_size; j++ )
   {
   // do not load AlreadyPaired (eg someone who was assigned a bye
      if( !AlreadyPaired[ player ] )
      {
          GroupPtr.group[ index ] = player;
          index++;
      }
      player++;
   }
   GroupPtr.rx_floats = rx_floats_count;
   group_size = index - 1;
   GroupPtr.count = group_size;
   return( group_size );
}


//////////////////////////////////////////////////////////////////
//
//  This function is called when pairing gets stuck.
//  It unpairs the previous group
//  and tries to pair all unpaired players as one combo group.
//  Keeps unpairing groups until pairing can be made or
//  all groups have been unpaired (pairing cannot be done)
//
int SwissCtrl_c::ComboPairing( int curr_group, int *group_size )
{
   int group;

   for( group = curr_group; group >= 0; group-- )
   {
 // we do 1 pass unnecessarily (group 0 does not exist but we handle the
 // situation when we are stuck on the first group
      if( group > 0 )
         UnpairGroup( group );
      *group_size = ComboBuildGroupVector( );
      PairGroup( &GroupPtr, TRUE );
      if( GroupPtr.count == (*group_size) / 2 )
      {
         MsgMngr->WriteToLog( LOG_MED, "ComboPairing()" );
         return( GroupPtr.count );
      }
   }
   return -1;
}


/******************** LOG UTILITIES ****************************/

void SwissCtrl_c::LogFinalPairs( int from_pair, int to_pair )
{
   int i;

   LogBuf[0] = 0;
   for( i = from_pair; i <= to_pair; i++ )
   {
      sprintf( LogBuf+strlen(LogBuf), "%3d-%d",
            FinalPairing[i].left, FinalPairing[i].right );
      if( strlen( LogBuf ) > 60 )
      {
         MsgMngr->WriteToLog( LOG_MED, LogBuf );
         LogBuf[0] = 0;
      }
   }
   MsgMngr->WriteToLog( LOG_MED, LogBuf );
}
