/*****************************************************************
*
* PROJECT:        Swiss Pairing Test Program
*
*                 Copyright (C) 1996 Robert Rozycki
*                 SwissPerfect/SPAR General Public Licence
*
* FILE:           sws_data.cpp
*
* DESCRIPTION:
*
* AUTHOR:         Robert Rozycki   rozycki@perth.dialix.oz.au
*
* PORTABILITY:    ANSI C++
*
* $Revision: 1.2 $
*
* $Date: 1996/02/04 09:43:48 $
*
* $Log: sws_data.cpp $
// Revision 1.2  1996/02/04  09:43:48  robert
// Internet release Feb 1996
//
// Revision 1.1  1996/01/01  14:04:35  robert
// Initial revision
//
*
*******************************************************************/

#include <stdio.h>
#include <stdlib.h>
//#include <mem.h>
#include "string.h"
#include "l_types.h"
#include "utils.h"
#include "sws_data.h"

#define PLAYERS_PER_VECTOR  200

////////////////////////////////////////////////////////////////////////
//
//  constructor
//
SwissData_c::SwissData_c( MsgMngr_c *msg_mngr ) : SwissGrPair_c( msg_mngr )
{
  ResetVars();
}

////////////////////////////////////////////////////////////////////////
//
//  destructor
//
SwissData_c::~SwissData_c()
{
    int i;

    for( i = 0; i < PlayersCount; i++ )
       free( AllOppons[ i ] );
    free( AllOppons );
    free( OpponsCount );
    free( UpFloats );
    free( DownFloats );
    free( Byes );
    free( ColPref );
}


/*====================== I N T E R F A C E ===========================*/


////////////////////////////////////////////////////////////////////////
//
//  Calls main pairing function and tidies up the pairing afterwards
//
int SwissData_c::MakePairing( PAIR *pairing, int players_count,
                              int round, int one_group)
{
   int result;

   result = SwissCtrl_c::MakePairing( pairing, players_count,
                                      round, one_group );
   SortRoundTab ( pairing );
   AssignColours( pairing );
   DemapRoundTab  ( pairing, players_count );
   return result;
}

/* ============= Interface functions for retrieving data =================*/


////////////////////////////////////////////////////////////////////////
//
//   Returns number of players the player has played against already
//      (forbidden opponents)
//
int SwissData_c::GetOpponCount( int player )
{
    if( player > 0 && player <= PlayersCount )
       return( OpponsCount[ player ] );
    return( 0 );
}

////////////////////////////////////////////////////////////////////////
//
//   Returns player's opponent start number for opponent index "idx"
//
int SwissData_c::GetOppon( int player, int idx )
{
   if( idx >= 0 && idx < OpponsCount[ player ] )
       return( AllOppons[ player ][ idx ] );
   return( 0 );
}

////////////////////////////////////////////////////////////////////////
//
//   This function is called when an application wants to
//   know what players are in a particular point group.
//   As players' numbers are renumbered (mapped) they are
//   consecutive within a group.  Thus it is enough to
//   return first and last number in the group.
//
//   PARAMETERS:  group  -  group number
//                first  -  pointer to first group player's start number
//                last   -  pointer to last group player's start number
//
//   RETURNS:     SUCCESS
//                FAILURE - if parameters inconsistent with the tournament
//                         data
//
int SwissData_c::GetGroup( int group, int& first, int& last )
{
   if( group > GroupsCount || GroupMarks[ group ] > PlayersCount )
      return( FAILURE );
   first = GroupMarks[ group ];
   if( group == GroupsCount )
       last = PlayersCount;
   else
       last = GroupMarks[ group + 1 ] - 1;
   return( SUCCESS );
}

////////////////////////////////////////////////////////////////////////
//
//    Returns TRUE if player1 has not played player 2 yet
//    or FALSE otherwise
//
//
//    NOTE:    It is important that player1's opponent list is checked
//             against the presence of player2 not the other way round.
//             Some applications may rely on that.
//
int SwissData_c::CanPlay( int player1, int player2 )
{
   int i;

   for( i = 0; i < OpponsCount[ player1 ]; i++ )
   {
      if( AllOppons[ player1 ][ i ] == player2 )
         return( FALSE );
   }
   return( TRUE );
}


/* ========= Interface functions for loading data =======================*/

/*
    NOTE:  In future these functions should not be here but in a derived
           object.  We may want to have data read in from different sources
           or in different formats so we will need different objects
           derived from SwissData_c handling data loading.
*/

//////////////////////////////////////////////////////////////////////
//
//   This function is called when it is discovered that
//   the player has played against a player who has not
//   been included in the original list of opponents.
//   The memory must be reallocated.
//
int *SwissData_c::ReallocOpponVector( int player, int new_count )
{
   int *temp_ptr;

   temp_ptr = AllOppons[ player ];
   AllOppons[ player ] = (int *)realloc( temp_ptr, new_count * sizeof(int) );
   if( AllOppons[ player ] == NULL )
       AllOppons[ player ] = temp_ptr;
   return( AllOppons[ player ] );
}

//////////////////////////////////////////////////////////////////////
//
//   Allocates memory to hold opponents list for the player.
//   If forbidden is set to FALSE then "oppon_count" indicates
//   number of allowed opponents rather forbidden ones (default).
//
int *SwissData_c::AllocOpponVector( int player, int oppon_count, int forbidden )
{
   int alloc_count;

   if( forbidden == FALSE )
       alloc_count = PlayersCount - oppon_count;
   else
       alloc_count = oppon_count;
   OpponsCount[ player ] = 0;
   AllOppons[ player ] = (int *)malloc( alloc_count * sizeof( int ) ) ;
   return( AllOppons[ player ] );
}

//////////////////////////////////////////////////////////////////////
//
//   Add's an opponent to the player's opponent list
//   and then checks if the player is on the opponent's list
//   as well.
//   If not corrects that.
//
int SwissData_c::LoadOppon( int player, int opp_no )
{
   int i, idx;

   idx = OpponsCount[ player ];
   AllOppons[ player ][ idx ] = opp_no;
   OpponsCount[ player ]++;

   if( opp_no > player )
      return( SUCCESS );

 /* check if "opp_no" has "player" as an opponent already */
   for( i = 0; i < OpponsCount[ opp_no ]; i++ )
   {
      if( AllOppons[ opp_no ][ i ] == player )
         return( SUCCESS );
   }

 /* allocate new vector */
   OpponsCount[ opp_no ]++;
   if( ReallocOpponVector( opp_no, OpponsCount[opp_no] ) == NULL )
       return FAILURE;
/*
   new_ptr = (int *)malloc( OpponsCount[ opp_no ] * sizeof( int ) ) ;
   if( new_ptr == NULL )
       return( FAILURE );
   memcpy( new_ptr, AllOppons[ opp_no ], OpponsCount[ opp_no ] * sizeof( int ) );
   free( AllOppons[ opp_no ] );
*/
   AllOppons[ opp_no ][ OpponsCount[ opp_no ] - 1 ] = player;
   return( SUCCESS );
}


//////////////////////////////////////////////////////////////////////
//
//  Initialises object's variables
//
void SwissData_c::ResetVars( void )
{
   MVector = 0;
   PlayersCount = 0;
   GroupsCount = 0;
   GroupMarks[ 0 ] = 1;
}

//////////////////////////////////////////////////////////////////////
//
//  Add new player to MapVectors and allocate new vector if necessary
//
int SwissData_c::AddPlayer( int player )
{
/* allocate next vector if needed */
   if( ( PlayersCount ) % PLAYERS_PER_VECTOR == 0 )
   {
       MVector++;
       if( ( MapVectors[ MVector ] =
             (int *)malloc( PLAYERS_PER_VECTOR * sizeof( int ) ) ) == NULL )
       {
          return( nomem_msg() );
       }
       else
       {
          memset( MapVectors[MVector], 0, PLAYERS_PER_VECTOR * sizeof(int) );
       }
   }
   MapVectors[ MVector ][ PlayersCount % PLAYERS_PER_VECTOR ] = player;
   PlayersCount++;
   return( SUCCESS );
}

//////////////////////////////////////////////////////////////////////
//
//   Finds the internal player's number corresponding
//   to his original one.
//
int SwissData_c::MapPlayer( int player, int map_option )
{
   int i, j, vectors_count;

   vectors_count = ( ( PlayersCount - 1 ) / PLAYERS_PER_VECTOR ) + 1;
   if( map_option == FALSE )
      return( player );
   if( player == 0 )
      return 0;
   for( i = 1; i <= vectors_count; i++ )
   {
      for( j = 0; j < PLAYERS_PER_VECTOR; j++ )
      {
         if( MapVectors[ i ][ j ] == player )
            return( ( i-1 ) * PLAYERS_PER_VECTOR + j + 1 );
      }
   }
   return( FAILURE );
}

//////////////////////////////////////////////////////////////////////
//
//   Finds the original player's start no.
//
int SwissData_c::DemapPlayer( int player, int map_option )
{
   int vector, offs;

   if( map_option == FALSE )
      return( player );
   if( player == 0 )
      return 0;
   vector = ( ( player - 1 ) / PLAYERS_PER_VECTOR ) + 1;
   offs   = ( player - 1 ) % PLAYERS_PER_VECTOR;

   return MapVectors[ vector ][ offs ];
}

//////////////////////////////////////////////////////////////////////
//
//   Maps pairing table
//
void SwissData_c::DemapRoundTab( PAIR *pairing, int players_count )
{
   int i, pairs;

   pairs = ( players_count >> 1 ) + ( players_count & 0x01 );
   for( i = 1; i <= pairs; i++ )
   {
      pairing[i].left  = DemapPlayer( pairing[i].left );
      pairing[i].right = DemapPlayer( pairing[i].right );
   }
}

//////////////////////////////////////////////////////////////////////
//
// Allocates memory for object's data structures
//
int SwissData_c::AllocTables()
{
   int trn_size, int_size, char_size, trntab_size, chartab_size;

/* calculate some useful values */
   trn_size     = PlayersCount + 1;
   int_size     = sizeof( int );
   char_size    = sizeof( char );
   trntab_size  = trn_size * int_size;
   chartab_size = trn_size * char_size;

/* allocate memory for pointers to opponent vectors */
   if(( AllOppons = (int **)malloc( trn_size * sizeof(int *))) == NULL )
      return( nomem_msg() );
   else
      memset( AllOppons, 0, trn_size * sizeof( int * ) );

/* allocate memory for opponent count (number of Oppons for each player */
   if( ( OpponsCount = (int *)malloc( trntab_size ) ) == NULL)
      return( nomem_msg() );
   else
      memset( OpponsCount, 0, trntab_size );

 /* allocate memory for upfloat status of each player */
   if( ( UpFloats = (char *)malloc( chartab_size ) ) == NULL )
      return( nomem_msg() );
   else
      memset( UpFloats, 0, chartab_size );

 /* allocate memory for downfloat status of each player */
   if( ( DownFloats = (char *)malloc( chartab_size ) ) == NULL)
      return( nomem_msg() );
   else
      memset( DownFloats, 0, chartab_size );

 /* allocate memory for Byes status of each player */
   if( ( Byes = (char *)malloc( chartab_size ) ) == NULL)
      return( nomem_msg() );
   else
      memset( Byes, 0, chartab_size );

 /* allocate memory for colour preference of each player */
   if( ( ColPref = (int *)malloc( trntab_size ) ) == NULL)
      return( nomem_msg() );
   else
      memset( ColPref, 0, trntab_size );
   return( SUCCESS );
}

//////////////////////////////////////////////////////////////////////
//
//   Marks the group end.
//   Last stored player is last in that group.
//
void SwissData_c::NewGroupMark( void )
{
   if( PlayersCount == 0 )
       return;
   GroupsCount++;
   GroupMarks[ GroupsCount ] = PlayersCount + 1;
}

