/*****************************************************************
*
* PROJECT:        Swiss Perfect 3
*
*                 Copyright (C) 1996 Robert Rozycki
*                 SwissPerfect/SPAR General Public Licence
*
* FILE:           sws_pair.cpp
*
* DESCRIPTION:    Pairs a point group.
*
* AUTHOR:         Robert Rozycki     rozycki@perth.dialix.oz.au
*
* PORTABILITY:    ANSI C++
*
* $Revision: 1.1 $
*
* $Date: 1996/02/04 09:51:23 $
*
* $Log: sws_pair.cpp $
// Revision 1.1  1996/02/04  09:51:23  robert
// Initial revision
//
*
*******************************************************************/

#include <string.h>
#include <stdlib.h>

#include "sws_pair.h"



/*========================= I N T E R F A C E ===========================*/

////////////////////////////////////////////////////////////////////
// Constructor
//
SwissGrPair_c::SwissGrPair_c( MsgMngr_c *msg_mngr )
             : SwissCtrl_c( msg_mngr )
{
   MsgMngr = msg_mngr;
   PairStat.allowed_ccl = BAD_COMPAT;
}

////////////////////////////////////////////////////////////////////
//
// DESCRIPTION:  Initialises object's data, retrieves group marks.
//               Loads back created pairs when finished.
// PARAMETERS:   pair_group -    pointer to a group (list of players)
//                               to be paired
//               swiss         - pointer to a SWISS object (tournament data)
//               is_last_group - indicates if we're pairing the last group
//
// RETURNS:      number of pairs made if successful
//               FAILURE    - if pairing impossible
//               OUT_OF_MEM - on out of memory condition
//
int SwissGrPair_c::PairGroup( PAIRED_VEC *pair_group, int is_last_group )
{
   int i, idx, pairs_made, with_no_oppons;

   IsLastGroup = is_last_group;
   if( pair_group->count <= 1 )
       return( FAILURE );   // should never happen
// initialise object data
   TotalGroupSize = pair_group->count;
   GroupSize  = TotalGroupSize;
   FixedPairs = 0;
   pGroup     = pair_group->group;
   RxFloats   = pair_group->rx_floats;
   if( AllocGroupPairingTables() != SUCCESS )
      return OUT_OF_MEM;
   with_no_oppons = RemoveNoHopers();
   if( ( with_no_oppons != 0 ) && ( IsLastGroup == TRUE ) )
   {
      if( with_no_oppons == 1 )
      {
         if( ( GroupSize & 0x01 ) == 0 )
            return( FAILURE );
      }
      else
         return( FAILURE );
   }
   GroupSize -= with_no_oppons;

   if( GroupSize <= 1 )   // no pairing can be done
   {
      pair_group->count = 0;
      return( SUCCESS );
   }
   else
      pairs_made = PairingControl( );

   for( i = 1; i <= pairs_made; i++ )
   {
      GrPairing[ i ].left  = pGroup[ GrPairing[ i ].left ];
      GrPairing[ i ].right = pGroup[ GrPairing[ i ].right ];
   }

// If GroupSize is odd then remove temporarily the unpaired player
// If there are any fixed pairs they'll be moved as required
// This is necessary for AppendFixedPairs() and AppendFloaters()
// to work correctly
   if( GroupSize & 0x01 )
   {
      RemovePlayer( GrPairing[ pairs_made+1 ].left, TotalGroupSize );
      GroupSize--;
   }

   AppendFixedPairs( pairs_made, GrPairing );
   pairs_made += FixedPairs;
   GroupSize += ( FixedPairs << 1 );
   AppendFloaters( pairs_made, GrPairing );
   i   = 0;
   idx = 1;
   while( TRUE )
   {
      pair_group->group[ idx++ ] = GrPairing[i+1].left;
      if( idx > GroupSize )
         break;
      pair_group->group[ idx++ ] = GrPairing[i+1].right;
      if( idx > GroupSize )
         break;
      i++;
   }
   pair_group->count = pairs_made;
   return( pairs_made );
}

/*==================== I M P L E M E N T A T I O N  ====================*/


/*=======================  PAIRING CONTROL ===========================*/

//////////////////////////////////////////////////////////////////////
//
//  Controls the entire process of group pairing.
//  Controls switching form one state to another
//  RETURNS: number of pairs made
//
int SwissGrPair_c::PairingControl( )
{
   int result;

   memset( &PairStat, 0, sizeof( PairingStatus_t ) );
   PairStat.state = NO_EXCHANGES_OPTIMAL;
   PairStat.max_noncols = CalcMaxNonCols();
   PairStat.allowed_ccl = BAD_COMPAT;

   while( TRUE )
   {
      switch( PairStat.state )
      {
         case NO_EXCHANGES_OPTIMAL:
            result = PairStateOptimal();
            if( result == SUCCESS )
               return GroupSize >> 1;
            else
               PairStat.state = FIND_ANY_PAIRING;
            break;
         case FIND_ANY_PAIRING:
            result = PairStateFindAny();
            if( result == SUCCESS )
            {
               SaveToBestPairing();
               PairStat.state = EXCHANGES_BEST_COLS;
            }
            else
               return FAILURE;
            break;
         case EXCHANGES_BEST_COLS:
            FixedPairs = RemoveFixedPairs();
            GroupSize  -= ( FixedPairs << 1 );
            if( GroupSize == 0 )
               result = SUCCESS;
            else
            {
               while( TRUE )
               {
                  result = PairStateExchBestCols();
                  if( result == SUCCESS )
                  {
                     SaveToBestPairing();
                     break;
                  }
                  else
                  {
                     if( RelaxColReqs() == FAILURE )
                     {
                        RetrieveFromBestPairing();
                        break;
                     }
                  }
               }   // while()
            }
            PairStat.state = NO_EXCHANGES_BEST;
            break;
         case NO_EXCHANGES_BEST:
            result = PairStateOptimal();
            if( result != SUCCESS )
               RetrieveFromBestPairing();
            return GroupSize >> 1;
      }
   }
}

//////////////////////////////////////////////////////////////////
//
// Tries to find optimal pairing which means:
//    - no exchanges
//    - maximum colour pairs
//    - best possible floater
//
int SwissGrPair_c::PairStateOptimal()
{
   int tx_float = 0;
   int rx_float_lev = 1;
   int pair_res;

// clear entire previous pairing
   PairStat.allow_exch = FALSE;
   InitSubGroupPairing( 1 );
   LoadRxFloats();
   while( TRUE )
   {
      if( PairRxFloats( rx_float_lev, PairStat.max_noncols ) != SUCCESS )
         return FAILURE;
      if( IsTxFloat )
         tx_float = SelectTxFloat( 0 );
      while( (pair_res = PairRemainder( 0, NO_EXCHANGES,
                            PairStat.max_noncols )) != SUCCESS )
      {
         if( IsTxFloat && PairStat.tx_float_stuck.stuck > 0 )
         {
            tx_float = SelectTxFloat( tx_float, TRUE );
            ClearBottomPairings( RxFloats+1, MaxPairs );
            ClearStuckData();
            if( tx_float == 0 )
            {
               tx_float = SelectTxFloat( tx_float, TRUE );
               break;
            }
            continue;
         }
         else
            break;
      }
      if( pair_res == SUCCESS )
         return SUCCESS;
      if( PairStat.rx_float_stuck.stuck == 0 &&
                PairStat.rx_float_oppon_stuck.stuck == 0 )
         return FAILURE;
      ClearBottomPairings( RxFloats+1, MaxPairs );
      ClearStuckData();
   }
}

//////////////////////////////////////////////////////////////////
//
//  Pairs as many pairs as legally possible.
//  Removes players who can't be paired or prevent others from
//  from being paired.
//  Does not care about colour pairs or exchanges.
//
int SwissGrPair_c::PairStateFindAny()
{
// clear entire previous pairing
   PairStat.allow_exch = TRUE;
   while( TRUE )
   {
      InitSubGroupPairing( 1 );
      LoadRxFloats();
      if( PairRxFloats( 1, MaxPairs ) != SUCCESS )
      {
         if( IsLastGroup )
            return FAILURE;
         RemoveStuckPlayer();
         continue;
      }
      InitSubGroupPairing( RxFloats+1 );
      if( PairRemainder( 0, ALLOW_EXCHANGES, MaxPairs )
                                != SUCCESS )
      {
         if( IsLastGroup )
            return FAILURE;
         RemoveStuckPlayer();
      }
      else
      {
         if( IsTxFloat )
            SelectTxFloat();
         return SUCCESS;
      }
   }
}

//////////////////////////////////////////////////////////////////
//
//  Finds pairing with required number of colour pairs.
//  Exchanges allowed.
//
int SwissGrPair_c::PairStateExchBestCols()
{
// clear entire previous pairing
   PairStat.allow_exch = TRUE;
   while( TRUE )
   {
      InitSubGroupPairing( 1 );
      LoadRxFloats();
      if( PairRxFloats( 1, PairStat.max_noncols ) != SUCCESS )
      {
         return FAILURE;
      }
      InitSubGroupPairing( RxFloats+1 );
      if( PairRemainder( 0, ALLOW_EXCHANGES, PairStat.max_noncols )
                                != SUCCESS )
      {
         return FAILURE;
      }
      else
      {
         if( IsTxFloat )
            SelectTxFloat();
         return SUCCESS;
      }
   }
   return SUCCESS;
}

//////////////////////////////////////////////////////////////////
//
//  Last stage of pairing. To Be Designed.
//
int SwissGrPair_c::PairStateNoExchBest()
{
   return SUCCESS;
}

//////////////////////////////////////////////////////////////////////
//
// Relaxes colour rules ie increases the number of non-colour pairs
// allowed.
//
int SwissGrPair_c::RelaxColReqs( )
{
   PairStat.max_noncols++;
   if( PairStat.max_noncols > MaxPairs )
      return FAILURE;
   else
      return SUCCESS;
}

//////////////////////////////////////////////////////////////////
//
//  Loads Rx Floaters into GrPairing[]
//
void SwissGrPair_c::LoadRxFloats()
{
   int i;

   for( i = 1; i <= RxFloats; i++ )
   {
      GrPairing[i].left = i;
      TempUsed[i] = TRUE;
   }
}

//////////////////////////////////////////////////////////////////////
//
//  Calculates maximum number of non-colour pairs allowed
//
int SwissGrPair_c::CalcMaxNonCols()
{
   int i, result, pref_white, pref_black, pref_none;
   int min_noncol_pairs;

   pref_black = pref_white = pref_none = 0;
   for( i = 1; i <= GroupSize; i++ )
   {
      result = GetColPref( pGroup[ i ] );
      if( result > 0 )
         pref_black++;
      else
      {
         if( result < 0 )
            pref_white++;
         else
            pref_none++;
      }
   }
   min_noncol_pairs = abs( pref_black - pref_white ) - pref_none;
   if( min_noncol_pairs <= 0 )
      min_noncol_pairs = 0;
   else
      min_noncol_pairs >>= 1;
   return( min_noncol_pairs );
}

//////////////////////////////////////////////////////////////////////
//
//  Calculates MidGroup ie finds the last player in the first half
//  of the group.
//  Players already paired are treated as removed.
//  This function is called after RxFloaters have been paired and
//  the remainder of the group is toi be paired.
//
int SwissGrPair_c::CalcMidGroup()
{
   int i, count;

   count = 0;
   i = 0;
   while( count < MaxPairs )
   {
      i++;
      if( TempUsed[i] == FALSE )
         count++;
   }
   return i;
}

//////////////////////////////////////////////////////////////////////
//
//  Replaces TxFloater with the next one depending on group's
//  and current floater's preferences
//
int SwissGrPair_c::SelectTxFloat( int curr_float, int same_qual )
{
   int i, search_start, search_col, tmp, qual;

   if( !IsTxFloat )
      return FAILURE;
   curr_float = GrPairing[MaxPairs+1].left;
   if( curr_float == 0 )
   {
      qual = 0;
      search_start = GroupSize;
      search_col = PairStat.prevail_col_pref;
   }
   else
   {
      qual = GetDnfStatus( pGroup[curr_float] );
      search_start = curr_float - 1;
      search_col = GetColPref( pGroup[curr_float] );
      if( search_col > 0 )
         search_col = 1;
      if( search_col < 0 )
         search_col = -1;
   }

   while( TRUE )
   {
      for( i = search_start; i > 0; i-- )
      {
         if( TempUsed[i] )
            continue;
         if( GetDnfStatus(pGroup[i]) == qual )
         {
            tmp = GetColPref( pGroup[i] );
            if( tmp > 0 )
               tmp = 1;
            if( tmp < 0 )
               tmp = -1;
            if( search_col == 0 || tmp == search_col )
            {
               TempUsed[curr_float] = FALSE;
               GrPairing[MaxPairs+1].left = i;
               TempUsed[i] = TRUE;
               return i;
            }
         }
      }
      if( same_qual )
         return 0;
      if( PairStat.prevail_col_pref != 0 &&
          search_col == PairStat.prevail_col_pref )
      {
         if( search_col > 1 )
            search_col = -1;
         else
            search_col = 1;
         search_start = GroupSize;
         continue;
      }
      if( qual < 2 )
      {
         qual++;
         search_start = GroupSize;
      }
      else
         return 0;
   }
}

//////////////////////////////////////////////////////////////////
//
// Pairs RxFloaters
//
int SwissGrPair_c::PairRxFloats( int modify_lev, int max_noncols )
{
// set MidGroup for pairing floaters
   MidGroup = RxFloats;
   return( PairSubGroup( 1, RxFloats, NO_EXCHANGES, modify_lev, max_noncols ) );
}

//////////////////////////////////////////////////////////////////
//
// Pairs Remainder Subgroup
//
int SwissGrPair_c::PairRemainder( int modify_lev, int allow_exch, int max_noncols )
{
   MidGroup = CalcMidGroup();
   if( !allow_exch )
       LoadLeftPlayers( RxFloats+1, MaxPairs );
   return( PairSubGroup( RxFloats+1, MaxPairs, allow_exch, modify_lev,
                             max_noncols ));
}

//////////////////////////////////////////////////////////////////
//
// Saves GrPairing to BestPairing
//
void SwissGrPair_c::SaveToBestPairing()
{
   int size;

   size = ( MaxPairs+1 ) * sizeof( PAIR_PARAMS );
   memcpy( BestPairing, GrPairing, size );
}

//////////////////////////////////////////////////////////////////
//
// Copies BestPairing to GrPairing
//
void SwissGrPair_c::RetrieveFromBestPairing()
{
   int size;

   size = ( MaxPairs+1 ) * sizeof( PAIR_PARAMS );
   memcpy( GrPairing, BestPairing, size );
}

//////////////////////////////////////////////////////////////////////
//
//  Initialises necessary data before PairSubGroup() is called
//
void SwissGrPair_c::InitSubGroupPairing( int clear_level )
{
   int tot_pairs;

   PairStat.stuck = 0;

   MaxPairs = GroupSize >> 1;
   if( ( MaxPairs << 1 ) == GroupSize )
      IsTxFloat = FALSE;
   else
      IsTxFloat = TRUE;
   MidGroup = CalcMidGroup();

   tot_pairs = ( TotalGroupSize >> 1 ) + ( TotalGroupSize & 0x01 );
   ClearBottomPairings( clear_level, tot_pairs );
}

//////////////////////////////////////////////////////////////////////
//
//  Clears GrPairing[] from 'clear_level' down to the end.
//  Unpairs encountered players. Does not touch TxFloater
//  if any.
//
void SwissGrPair_c::ClearBottomPairings( int clear_level,
                                         int bot_level, int both )
{
   int i;

//   UnpairLevels( clear_level, bot_level );
   for( i = clear_level; i <= bot_level; i++ )
   {
      if( both )
         GrPairing[i].left  = 0;
      GrPairing[i].right = 0;
   }
   memset( TempUsed, 0, (TotalGroupSize+1) * sizeof( int ) );
   for( i = 1; i <= bot_level; i++ )
   {
      TempUsed[ GrPairing[i].left ] = TRUE;
      TempUsed[ GrPairing[i].right ] = TRUE;
   }
}

/*===================== PAIRING SUBGROUPS ==========================*/

//////////////////////////////////////////////////////////////////////
//
//  Control function for pairing a subgroup.
//  Can be called by PairingControl() ONLY.
//
int SwissGrPair_c::PairSubGroup( int top_level, int bot_level,
                            int allow_exch, int modify_lev, int max_noncols )
{
   int level, prev_level, non_cols;
   int old_left = 0, new_left = 0, old_right = 0, new_right = 0;
   int pair_undone = -1;
   int strict_cols = FALSE;
   LevelSide_t lev_oppon;

// Initialise
   level = prev_level = top_level;
   lev_oppon.oppon = 0;

   if( modify_lev != 0 )
   {
      level = prev_level = modify_lev;
      pair_undone = RIGHT;
      ClearBottomPairings( level+1, MaxPairs, FALSE );
   }
   else
   {
      level = prev_level = top_level;
   }
   non_cols = CountNonCols( level-1 );
   while( TRUE )
   {
      if( level > bot_level || level < top_level )
         break;
      old_left  = GrPairing[ level ].left;
      old_right = GrPairing[ level ].right;

// Check if this is straight "after stuck" situation
      if( prev_level > level )
         non_cols = CountNonCols( level-1 );
      if( non_cols == max_noncols )
         strict_cols = TRUE;
      else
         strict_cols = FALSE;
      prev_level = level;

// find if "left" player if missing or removed
      if( old_left == 0 )
         new_left = GetNextFree();
      else
      {
         if( pair_undone == LEFT )  // straight after ReplaceWith...
         {
            new_left = GetNextFree( old_left );
            pair_undone = -1;
            // unpair right
            GrPairing[ level ].right = 0;
            TempUsed[ old_right ] = 0;
            old_right = 0;
         }
         else
            new_left = old_left;
      }
      GrPairing[level].left = new_left;
      TempUsed[new_left] = TRUE;
      if( new_left == 0 )
         return FAILURE;

// find "right" player
// skip this stage if right exists and pair_undone == -1
      if( old_right && pair_undone == -1 )
      {
         new_right = old_right;
      }
      else
      {
         if( level > RxFloats )
            new_right = GetNextOppon( level, strict_cols );
         else
            new_right = GetNextRxOppon( level, strict_cols );
         if( pair_undone == RIGHT )    // straight after ReplaceWith...
            pair_undone = -1;
         else
            TempUsed[old_right] = FALSE;
         GrPairing[level].right = new_right;
         TempUsed[new_right] = TRUE;
      }

      if( new_right == 0 )
      {
      // Update stuck info
         if( new_left > PairStat.stuck )
            PairStat.stuck = new_left;
         if( level > RxFloats )
            WouldModifyFloatsHelp( new_left, level, strict_cols );
         lev_oppon = ReplaceWithOpponFromAbove( level, strict_cols,
                              allow_exch, top_level );
         if( lev_oppon.level == 0 )  // stuck for good
         {
            return FAILURE;
         }
         else
         {
            pair_undone = lev_oppon.side;
            UnpairLevels( lev_oppon.level+1, level-1 );
            GrPairing[level].right = lev_oppon.oppon;
            level = lev_oppon.level;
         }
         continue;
      }
      if( ColourCompatLevel( new_left, new_right ) != FULL_COMPAT )
      {
         non_cols++;
         if( non_cols > max_noncols )
         {
            UnpairLevels( level, level );
            level--;
            non_cols--;
         }
      }
      level++;
      if( PairStat.stuck_level != 0 && level > PairStat.stuck_level )
         ClearStuckData();
   }
   if( level < top_level )
      return FAILURE;
   else
      return SUCCESS;
}

//////////////////////////////////////////////////////////////////////
//
// Finds the next player in the group not paired yet and with the
// "ord" number > *current*
//
int SwissGrPair_c::GetNextFree( int current )
{
   int i;

   for( i = current+1; i <= GroupSize; i++ )
   {
      if( TempUsed[ i ] == FALSE )
         return( i );
   }
   return 0;
}

//////////////////////////////////////////////////////////////////////
//
//  Find next legal opponent for "left" player at 'level' in
//  GrPairing[]. If strict_cols is TRUE then the opponent must
//  have compatible colour preference
//
int SwissGrPair_c::GetNextOppon( int level, int strict_cols )
{
   int i, left, right, cl, start;

   left  = GrPairing[level].left;
   right = GrPairing[level].right;
   if( !PairStat.allow_exch )  // left side is fixed
   {
      for( i = right+1; i <= GroupSize; i++ )
      {
         if( TempUsed[i] )
            continue;
         cl = CompatLevel( left, i );
         if( cl == FULL_COMPAT ||
                ( cl != INCOMPATIBLE && !strict_cols ) )
            return i;
      }
      return 0;
   }
// exchanges allowed
   if( right > MidGroup || right == 0 )
   {
      if( right == 0 )
         start = MidGroup+1;
      else
         start = right+1;
      for( i = start; i <= GroupSize; i++ )
      {
         if( TempUsed[i] )
            continue;
         cl = CompatLevel( left, i );
         if( cl == FULL_COMPAT ||
                ( cl != INCOMPATIBLE && !strict_cols ) )
            return i;
      }
   }

   if( right > MidGroup || right == 0 )
      start = MidGroup;
   else
      start = right - 1;

   for( i = start; i > 0; i-- )
   {
      if( TempUsed[i] )
         continue;
      cl = CompatLevel( left, i );
      if( cl == FULL_COMPAT ||
             ( cl != INCOMPATIBLE && !strict_cols ) )
         return i;
   }
   return 0;
}


//////////////////////////////////////////////////////////////////////
//
//  Find next legal opponent for RxFloater at 'level' in
//  GrPairing[]. If strict_cols is TRUE then the opponent must
//  have compatible colour preference
//
int SwissGrPair_c::GetNextRxOppon( int level, int strict_cols )
{
   int i, left, right, cl;
   int curr_qual;

   right = GrPairing[level].right;
   if( right == 0 )
      curr_qual = 0;
   else
      curr_qual = GetUpfStatus( pGroup[ right ] );
   left  = GrPairing[level].left;
//avoid consecutive upfloats if possible
   while( curr_qual <= 2 )
   {
      for( i = right+1; i <= GroupSize; i++ )
      {
         if( TempUsed[ i ] )
            continue;
         cl = CompatLevel( left, i );
         if( cl == FULL_COMPAT || ( cl != INCOMPATIBLE && !strict_cols ) )
            if( GetUpfStatus( pGroup[ i ] ) == curr_qual )
               return i;
      }
      curr_qual++;
   }
   return 0;
}


//////////////////////////////////////////////////////////////////////
//
//  The pairing process got stuck at *level* (no opponent for white there)
//  and this routine was called to find if there is a player paired
//  before who can play the "stuck" white.
//  If *allow_exch* is FALSE then only "right" players are tested
//  otherwise "right" players first and then "left" ones.
//  *top_level* is a level up to which the checking is done.
//  This should be the first pair in the homogeneous subgroup and the
//  very first pair if we are pairing floaters.
//
LevelSide_t SwissGrPair_c::ReplaceWithOpponFromAbove( int level,
                    int strict_cols, int allow_exch, int top_level )
{
   int i, left, right;
   LevelSide_t level_side;
   int compat_level;

   left = GrPairing[level].left;
   level_side.level = 0;
   level_side.side  = -1;
   level_side.oppon = 0;
   for( i = level - 1; i >= top_level; i-- )
   {
      right = GrPairing[i].right;
      if( right > 0 && CompatLevel( left, right ) != INCOMPATIBLE )
      {
         level_side.level = i;
         level_side.side = RIGHT;
         level_side.oppon = right;
         return level_side;
      }
   }
   if( allow_exch )
      return level_side;

   for( i = level - 1; i >= top_level; i-- )
   {
      right = GrPairing[i].left;
      compat_level = CompatLevel( left, right );
      if( right > 0 && (compat_level == FULL_COMPAT ||
         ( compat_level != INCOMPATIBLE && !strict_cols ) ) )
      {
         level_side.level = i;
         level_side.side = LEFT;
         level_side.oppon = right;
         return level_side;
      }
   }
   return level_side;
}

//////////////////////////////////////////////////////////////////////
//
//  Loads "left" players between *top_level* and *bot_level* inclusive
//
void SwissGrPair_c::LoadLeftPlayers( int top_level, int bot_level )
{
   int i, ind;

   i = top_level;
   ind = 1;
   while( i <= bot_level )
   {
      while( TempUsed[ind] )
         ind++;
      TempUsed[ind] = TRUE;
      GrPairing[i].left = ind;
      i++;
   }
}

//////////////////////////////////////////////////////////////////////
//
//  Unpairs all pairs between *top_level* and *bot_level* inclusive
//
void SwissGrPair_c::UnpairLevels( int top_level, int bot_level )
{
   int i;

   for( i = top_level; i <= bot_level; i++ )
   {
      if( PairStat.allow_exch )
      {
         TempUsed[ GrPairing[i].left ] = FALSE;
         GrPairing[i].left = 0;
      }
      TempUsed[ GrPairing[i].right ] = FALSE;
      GrPairing[i].right = 0;
   }
}

//////////////////////////////////////////////////////////////////////
//
//  This function is called when pairing gets stuck.  It tries to
//  establish if the pairing may get unstuck if we change RxFloaters
//  pairing or select another TxFloater
//
void SwissGrPair_c::WouldModifyFloatsHelp( int ord, int level, int strict_cols )
{
   int i, res;

   if( IsTxFloat && GrPairing[MaxPairs+1].left != 0 )
   {
      res = CompatLevel( GrPairing[MaxPairs+1].left, ord );
      if( res == FULL_COMPAT || !strict_cols && res <= PairStat.allowed_ccl )
      {
         if( PairStat.tx_float_stuck.stuck < ord )
         {
            PairStat.tx_float_stuck.stuck = ord;
            PairStat.tx_float_stuck.level = level;
         }
         return;
      }
   }

   for( i = RxFloats; i >=1; i-- )
   {
      res = CompatLevel( GrPairing[i].right, ord );
      if( res == FULL_COMPAT || !strict_cols && res <= PairStat.allowed_ccl )
      {
         if( PairStat.rx_float_oppon_stuck.stuck < ord )
         {
            PairStat.rx_float_oppon_stuck.stuck = ord;
            PairStat.rx_float_oppon_stuck.level = level;
         }
         return;
      }
   }

   for( i = RxFloats; i >=1; i-- )
   {
      res = CompatLevel( GrPairing[i].left, ord );
      if( res == FULL_COMPAT || !strict_cols && res <= PairStat.allowed_ccl )
      {
         if( PairStat.rx_float_stuck.stuck < ord )
         {
            PairStat.rx_float_stuck.stuck = ord;
            PairStat.rx_float_stuck.level = level;
         }
         return;
      }
   }
   return;
}

//////////////////////////////////////////////////////////////////////
//
// Clears "stuck info" in PairStat.
//
void SwissGrPair_c::ClearStuckData()
{
   PairStat.tx_float_stuck.stuck = 0;
   PairStat.rx_float_stuck.stuck = 0;
   PairStat.rx_float_oppon_stuck.stuck = 0;

   PairStat.stuck       = 0;
   PairStat.stuck_level = 0;
   PairStat.stuck_count = 0;
}

//////////////////////////////////////////////////////////////////////
//
//  Counts non-colour pairs between level 1 and 'level'
//
int SwissGrPair_c::CountNonCols( int level )
{
   int i, non_col;

   non_col = 0;
   for( i = 1; i <= level; i++ )
   {
      if( ColourCompatLevel( GrPairing[i].left, GrPairing[i].right )
                                       != FULL_COMPAT )
          non_col++;
   }
   return non_col;
}

//////////////////////////////////////////////////////////////////////
//
//  Establishes compatibility of two players.
//  First checks if they haven't played each other yet and then
//  their colour compatibility
//
int SwissGrPair_c::CompatLevel( int ord1, int ord2 )
{
   int ccl;

// If players cannot play each other at all return INCOMPATIBLE
   if( CanPlay( pGroup[ord1], pGroup[ord2] ) == FALSE )
      return( INCOMPATIBLE );
// get colour compatibility level (ccl)
   ccl = ColourCompatLevel( ord1, ord2 );
   if( ccl > PairStat.allowed_ccl )
      return( INCOMPATIBLE );
   else
      return( ccl );
}

//////////////////////////////////////////////////////////////////////
//
//  Determines how compatible two players are colourwise.
//  RETURNS:    Compatibility level
//               0 (FULL_COMPAT) - fully compatible
//               1 - only mild preferences violated
//               2 - strong preference violated
//               3 - absolute preference violated
//               4 - and over absolute preference violated
//
int SwissGrPair_c::ColourCompatLevel( int ord1, int ord2 )
{
   int pref1, pref2;

   pref1 = GetColPref( pGroup[ ord1 ] );
   pref2 = GetColPref( pGroup[ ord2 ] );
 // If any player has no preference, compatibility is full
   if( pref1 == 0 || pref2 == 0 )
      return( FULL_COMPAT );
 // If players have opposite preferences, compatibility is full
   if( pref1 > 0 && pref2 < 0 || pref1 < 0 && pref2 > 0 )
      return( FULL_COMPAT );
 // get absolute values of both preferences
   pref1 = abs( pref1 );
   pref2 = abs( pref2 );
 // The violation level is equal to the lesser abs(preference)
   if( pref1 < pref2 )
      return( pref1 );
   else
      return( pref2 );
}


//////////////////////////////////////////////////////////////////////
//
// Checks if any of not paired players had a bye before
//
int SwissGrPair_c::MultipleByes( )
{
   int i;

   for( i = 1; i <= TotalGroupSize; i++ )
   {
      if( TempUsed[ i ] == FALSE )
         if( GetByeStatus( pGroup[ i ] ) != 0 )
            return( TRUE );
   }
   return( FALSE );
}


/*================= REMOVING PLAYERS FROM THE GROUP =====================*/

//////////////////////////////////////////////////////////////////
//
// DESCRIPTION:  Moves a given player (pos) to the end of the Group
//               (group_end)shifting all behind him left by one beforehand.
//               Used to "remove" non-pairable players from the Group.
//
void SwissGrPair_c::RemovePlayer( int pos, int group_end )
{
   int i, tmp;

   if( pos <= 0 )
      return;
   tmp = pGroup[ pos ];
   if( pos <= RxFloats )
       RxFloats--;
   for( i = pos; i <= group_end; i++ )
       pGroup[ i ] = pGroup[ i+1 ];
   pGroup[ group_end ] = tmp;
}


//////////////////////////////////////////////////////////////////
//
// Identify players who have one opponent in the Group
// and then put them at the end of the group together with
// their only possible opponent (they form a 'fixed pair').
// RETURNS number of fixed pairs made
//
int SwissGrPair_c::RemoveFixedPairs( )
{
   int i, j, oppons_count, oppon_no;
   int count = 0;

   i = 1;
   while( i <= GroupSize - count )
   {
      oppons_count = 0;
      for( j = 1; j <= GroupSize - count; j++ )
      {
         if(( i!=j ) && (CompatLevel( i, j ) != INCOMPATIBLE ))
         {
            oppons_count++;
            if( oppons_count == 1 )
               oppon_no = j;
            else
               break;
         }
      }
      if( oppons_count == 1 )
      {
      // remove higher number first !!
         if( i > oppon_no )
         {
            RemovePlayer( i, GroupSize );
            RemovePlayer( oppon_no, GroupSize );
         }
         else
         {
            RemovePlayer( oppon_no, GroupSize );
            RemovePlayer( i, GroupSize );
         }
         count += 2;
      }
      i++;
   }
   return( count >> 1 );
}

//////////////////////////////////////////////////////////////////
//
// Identify and remove players who have no opponents in the Group
//
int SwissGrPair_c::RemoveNoHopers( )
{
   int i, j, found;
   int count = 0;

 //### NOTE that RxFloats may change
   i = 1;
   while( i <= GroupSize - count )
   {
      found = FALSE;
      j = 1;
      while( j <= GroupSize - count )
      {
         if(( i!=j ) && (CompatLevel( i, j ) != INCOMPATIBLE ))
         {
            found = TRUE;
            i++;
            break;
         }
         j++;
      }
      if( found == FALSE )
      {
         RemovePlayer( i, GroupSize );
         count++;
      }
   }
   return( count );
}

//////////////////////////////////////////////////////////////////
//
// Remove an RxFloater who makes the pairing impossible ("stuck")
//
void SwissGrPair_c::RemoveRxFloatFromGroup( int ord )
{
   RemovePlayer( ord, GroupSize );
   PairStat.rx_float_again++;
   GroupSize--;
}

//////////////////////////////////////////////////////////////////
//
// Remove a player who makes the pairing impossible ("stuck")
//
void SwissGrPair_c::RemoveStuckPlayer( )
{
   if( PairStat.stuck > 0 )
      RemovePlayer( PairStat.stuck, GroupSize );
   GroupSize--;
}


/*===================  NON_PAIRING UTILITIES  =======================*/

//////////////////////////////////////////////////////////////////
//
//  Appends 'fixed pairs' at the end of the pairing table.
//
//  pairs_made - number of pairs in the table
//  pair_table - pointer to the pairing table
//
void SwissGrPair_c::AppendFixedPairs( int pairs_made, PAIR_PARAMS *pair_table )
{
   int i, player_no;

   for( i = 1; i <= FixedPairs; i++ )
   {
      player_no = GroupSize + ( i << 1 ) - 1;
      pair_table[ pairs_made+i ].left  = pGroup[ player_no ];
      TempUsed[ player_no ] = TRUE;
      pair_table[ pairs_made+i ].right  = pGroup[ player_no + 1 ];
      TempUsed[ player_no + 1 ] = TRUE;
   }
}

//////////////////////////////////////////////////////////////////
//
//  Finds non-paired players and appends them at the end of the
//  pairing table.
//
//  pairs_made - number of pairs in the table
//  pair_table - pointer to the pairing table
//
void SwissGrPair_c::AppendFloaters( int pairs_made, PAIR_PARAMS *pair_table )
{
   int i, ind;
   int tot_tables;

   tot_tables = ( TotalGroupSize >> 1 ) + ( TotalGroupSize & 0x01 );
   ind = (GroupSize & ~0x0001) + 1;
   for( i = pairs_made + 1; i <= tot_tables; i++ )
   {
      pair_table[ i ].left  = pGroup[ ind ];
      ind++;
      if( ind <= TotalGroupSize )
         pair_table[ i ].right = pGroup[ ind ];
      else
         pair_table[ i ].right = 0;
      ind++;
   }
}

/*================= MEMORY ALLOCATION FUNCTIONS ==================*/

//////////////////////////////////////////////////////////////////////
//
//  Allocates all necessary memory. Returns SUCCESS or OUT_OF_MEM
//
int SwissGrPair_c::AllocGroupPairingTables()
{
   int half_size;

   TempUsed = NULL;
   GrPairing = NULL;
   BestPairing = NULL;
   half_size = ( TotalGroupSize >> 1 ) + 2;

   TempUsed = (int *)malloc( ( TotalGroupSize + 1 ) * sizeof( int ));
   if( TempUsed == NULL )
   {
      FreeGroupPairingTables();
      return( OUT_OF_MEM );
   }
   memset( TempUsed, 0, (TotalGroupSize+1) * sizeof( int ) );

   GrPairing = (PAIR_PARAMS *)malloc( half_size * sizeof( PAIR_PARAMS ) );
   if( GrPairing == NULL )
   {
      FreeGroupPairingTables();
      return( OUT_OF_MEM );
   }
   memset( GrPairing, 0, half_size * sizeof( PAIR_PARAMS ) );

   BestPairing = (PAIR_PARAMS *)malloc( half_size * sizeof( PAIR_PARAMS ) );
   if( BestPairing == NULL )
   {
      FreeGroupPairingTables();
      return( OUT_OF_MEM );
   }
   memset( BestPairing, 0, half_size * sizeof( PAIR_PARAMS ) );
   return( SUCCESS );
}

//////////////////////////////////////////////////////////////////////
//
//  Blanks memory allocated by AllocTables
//
void SwissGrPair_c::ClearGroupPairingTables()
{
   int half_size;

   half_size = ( TotalGroupSize >> 1 ) + 2;

   memset( TempUsed,    0, ( GroupSize+1 ) * sizeof( int ) );
   memset( GrPairing,   0, ( half_size )  * sizeof( PAIR_PARAMS ) );
   memset( BestPairing, 0, ( half_size )  * sizeof( PAIR ) );
}

//////////////////////////////////////////////////////////////////////
//
//  Frees memory allocated by AllocTables
//
void SwissGrPair_c::FreeGroupPairingTables()
{
   free( TempUsed );
   free( GrPairing );
   free( BestPairing );
}


