/*****************************************************************
*
* PROJECT:        Swiss Pairing Test Program
*
*                 Copyright (C) 1996 Robert Rozycki
*                 SwissPerfect/SPAR General Public Licence
*
* FILE:           sws_read.cpp
*
* DESCRIPTION:    Reads Swiss data from a text file
*
* AUTHOR:         Robert Rozycki    rozycki@perth.dialix.oz.au
*
* PORTABILITY:    ANSI C++
*
* $Revision: 1.2 $
*
* $Date: 1996/02/04 09:44:46 $
*
* $Log: sws_read.cpp $
// Revision 1.2  1996/02/04  09:44:46  robert
// Internet release Feb 1996
//
// Revision 1.1  1996/01/01  14:04:57  robert
// Initial revision
//
*
*******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "l_types.h"
#include "spar.h"
#include "utils.h"
#include "sws_read.h"

/////////////////////////////////////////////////////////////////////////
//
//  Constructor
//
SwissRead_c::SwissRead_c( MsgMngr_c *msg_mngr ) : SwissData_c ( msg_mngr )
{
   MapOption = TRUE;
   MsgMngr = msg_mngr;
}

/////////////////////////////////////////////////////////////////////////
//
//  Controls reading the data file in two passes.
//  First pass establishes the number of players and creates mapping
//  vectors.
//  Second pass reads the file line by line and interprets data
//
//  NOTE: Data file must be successfully opened before calling
//        this procedure.
//
//  Returns SUCCESS or FAILURE if data is incorrect
//
int SwissRead_c::ReadData( FILE *fdata )
{
   int tmp;

   FirstPass( fdata );
 /* allocate memory */
   if( AllocTables() != SUCCESS )
      return( FAILURE );
 /* allocate memory for temporary opponent vector */
   tmp = GetPlayersCount() + 1;
   if( ( PlayerOppons = (int *) malloc( tmp * sizeof(int) ) ) == NULL )
   {
      nomem_msg();
      exit( 1 );
   }
   else
      memset( PlayerOppons, 0, tmp * sizeof(int) );

   return( SecondPass( fdata ) );
}

//////////////////////////////////////////////////////////////////////
//
//  Reads the data file to establish the number of players
//  and then creates mapping vectors.
//
int SwissRead_c::FirstPass( FILE *fdata )
{
   int line_no, player;
   int max_player = 0;
   int group_count = 1;

   line_no = 0;
   while( fgets( LineBuf, MAX_SWSLINE, fdata ) != NULL )
   {
      line_no++;
      player = GetPlayer( LineBuf );
      if( player == 0 )
      {
         if( GetLineType( LineBuf ) == GROUP_LINE )
         {
            NewGroupMark();
            group_count++;
         }
         continue;
      }
      if( player == -1 )
         return( FAILURE );
      if( player > max_player )
         max_player = player;
      if( AddPlayer( player ) != SUCCESS )
         exit( 1 );
   }
   return( SUCCESS );
}

//////////////////////////////////////////////////////////////////////
//
//  Reads the data file line by line and interprets data
//
int SwissRead_c::SecondPass( FILE *fdata )
{
   int line_no, line_ret;

 /* go to the beginning of the file */
   fseek( fdata, 0L, SEEK_SET );
   line_no = 0;
   while( fgets( LineBuf, MAX_SWSLINE, fdata ) != NULL )
   {
      line_no++;
      line_ret = ExamineLine( line_no, LineBuf );
      switch( line_ret )
      {
         case SUCCESS:
           break;
         case FAILURE:
           return( FAILURE );
         case GROUP_LINE:
           break;
         default:
           print_error("Unknown ERROR", line_no );
      }
   }
  return( SUCCESS );
}

//////////////////////////////////////////////////////////////////////
//
//  Determines what type of line we've just read
//
int SwissRead_c::GetLineType( char *p )
{
   char buf[ MAX_TOKEN + 1 ];
   int count;

   p = skip_blanks( p );
   if( isdigit( *p ) )
      return( PLAYER_LINE );
   if( *p == '#' || *p == '/' || *p == '*' )
      return( COMMENT_LINE );
   count = 0;
   while( isalpha( *p ) && count < MAX_TOKEN )
   {
      buf[ count++ ] = *p;
      p++;
   }
   buf[ count ] = '\0';
   upper_str( buf );
   if( strcmp( buf, "GROUP" ) == 0 )
      return( GROUP_LINE );
   if( *p == '\0' )
      return( EMPTY_LINE );
   else
      return( UNKNOWN_LINE );
}

//////////////////////////////////////////////////////////////////////
//
//  Moves along a line (string) with player data
//  and calls other function to interpret encountered data
//
int SwissRead_c::ExamineLine( int line_no, char *p )
{
   int no, player;

/* read player number */
   no = GetPlayer( p );
   if( no == 0 )
   {
      switch ( GetLineType( p ) )
      {
         case COMMENT_LINE:
            return( SUCCESS );
         case EMPTY_LINE:
            return( SUCCESS );
         case GROUP_LINE:
            return( SUCCESS );
         default:
            return( FAILURE );
      }
   }
   player = MapPlayer( no, MapOption );
   p = goto_separ( p, SEPAR );
   p++;
   if( GetOpponents( p, player ) == NULL )
      return( FAILURE );
   while( *(p = goto_separ( p, SEPAR )) == SEPAR )
   {
      p++;
      p = skip_blanks( p );
      switch( *p )
      {
         case 'W':
         case 'w':
         case 'B':
         case 'b':
         case 'N':
         case 'n':
         case 'P':
         case 'p':
           p = GetColpref( player, p, line_no );
           break;
         case 'U':
         case 'u':
         case 'D':
         case 'd':
           p = GetFloats( player, p );
           break;
         default:
           print_error( "UNRECOGNIZED ENTRY", line_no );
           break;
      }
   }
   return( SUCCESS );
}

////////////////////////////////////////////////////////////////////////
//
//   Extracts player's start number from the line
//
int SwissRead_c::GetPlayer( char *p )
{
   int player;

   p = skip_blanks( p );
   if( !isdigit( *p ) )
      return( 0 );
   player = atoi( p );
   return( player );
}


//////////////////////////////////////////////////////////////////////
//
//    If PlayerOppons[] contains forbidden opponents
//    this function takes entries from that array
//    and calls swiss.LoadOppon[] to store them in the
//    SwissData_c object.
//    If PlayerOppons[] contains allowed opponents
//    rather than the forbidden ones this function takes
//    all players which are NOT in PlayerOppons[].
//
int SwissRead_c::StoreOppons( int player, int oppon_count, int forbidden )
{
   int i, j, found;
   int extra_oppons = 0;

   if( oppon_count != 0 || forbidden == FALSE )
   {
      if( AllocOpponVector( player, oppon_count, forbidden ) == NULL )
         return( nomem_msg() );
      if( forbidden == TRUE )
      {
         for( i = 0; i < oppon_count; i++ )
            if( LoadOppon( player, PlayerOppons[ i ] ) != SUCCESS )
               return( nomem_msg() );
      }
      else
      {
          for( i = 1; i <= GetPlayersCount(); i++ )
          {
             found = FALSE;
             for( j = 0; j < oppon_count; j++ )
             {
                if( i == PlayerOppons[ j ] )
                {
                   found = TRUE;
                   break;
                }
             }
             if( found == TRUE )
                found = FALSE;
             else
             {
                if( LoadOppon( player, i ) != SUCCESS )
                   return( nomem_msg() );
             }
          }
       }
   }
 // Check if this player was an opponent of someone above (lower start no )
 // and does not have that player as an opponent for himself in which case
 // we have to that player to this player's list

 // First see how many extra opponents we've got
   extra_oppons = 0;
   for( i = 1; i < player; i++ )
   {
 // NOTE: this is where we rely on the fact that CanPlay() looks for
 //       its second argument on its first argument's list.
 //       See the corresponding comment in SwissData_c::CanPlay()
      if( CanPlay( i, player ) == FALSE &&
               CanPlay( player, i ) == TRUE )
         extra_oppons++;
   }
   if( extra_oppons == 0 )
      return( SUCCESS );
 // Apply our clumsy procedure of reallocating memory
   if( ReallocOpponVector( player, oppon_count + extra_oppons ) == NULL )
      return( nomem_msg() );
 // Repeat the loop as above but this time store new opponents
   for( i = 1; i < player; i++ )
   {
      if( CanPlay( i, player ) == FALSE &&
         CanPlay( player, i ) == TRUE )
      {
         if( LoadOppon( player, i ) != SUCCESS )
            return( nomem_msg() );
      }
   }
   return( SUCCESS );
}

//////////////////////////////////////////////////////////////////////
//
//  Checks if an opponent is specified for the second time.
//  It is essential not to have same opponents twice in
//  PlayerOppons[] for the sake of StoreOppons() procedure
//  which may fail when we deal with allowed opponents
//  rather than forbiden ones.
//  Subtracting allowed opponents from the total number of
//  players will give an incorrect result and not enough
//  memory will be allocated for AllOppons[][].
//
// PARAMETERS:   oppon   - opponent's number
//               opp_idx - number of opponents so far
//
int SwissRead_c::NewOppon( int oppon, int opp_idx )
{
   int i;

   for( i = 0; i < opp_idx; i++ )
      if( PlayerOppons[ i ] == oppon )
         return( FALSE );
   return( TRUE );
}

//////////////////////////////////////////////////////////////////////
//
//   Reads opponents and their status (forbidden or allowed)
//   from the line string
//
char *SwissRead_c::GetOpponents( char *p, int player )
{
   int i, first_oppon, last_oppon;
   int forbidden = TRUE;
   int oppon_idx = 0;
   int oppon;

   p = skip_blanks( p );
   if( *p == 'A' )
   {
      forbidden = FALSE;
      p++;
   }
   while( TRUE )
   {
       p = skip_blanks( p );
       if( !isdigit( *p ) || *p == SEPAR )
       {
          if( StoreOppons( player, oppon_idx, forbidden ) != SUCCESS )
             return( NULL );
          return( p );
       }
       first_oppon = atoi( p );
       last_oppon = 0;
       p = skip_digits( p );
       p = skip_blanks( p );
       if( *p == '-' )
       {
          p++;
          p = skip_blanks( p );
          last_oppon = atoi( p );
          p++;
       }
       oppon = MapPlayer( first_oppon, MapOption );
       if( oppon > 0 && NewOppon( oppon, oppon_idx ) )
       {
          PlayerOppons[ oppon_idx ] = oppon;
          oppon_idx++;
       }
       if( last_oppon != 0 )
       {
          for( i = first_oppon + 1; i <= last_oppon; i++ )
          {
             oppon = MapPlayer( i, MapOption );
             if( oppon > 0 && NewOppon( oppon, oppon_idx ) )
             {
                PlayerOppons[ oppon_idx ] = oppon;
                oppon_idx++;
             }
          }
       }
    }
}

//////////////////////////////////////////////////////////////////////
//
//   Determines colour preference for a player
//
char *SwissRead_c::GetColpref( int player, char *p, int line_no )
{
   int col_balance    = 0;
   int last_col       = 0;
   int secondlast_col = 0;

   while( isalpha( *p ) )
   {
      switch( *p )
      {
        case 'N':
        case 'n':
        case 'X':
        case 'x':
        case '-':
        case '_':
          break;
        case 'P':
        case 'p':
          LoadByeStatus( player, 1 );
          break;
        case 'W':
        case 'w':
          col_balance++;
          secondlast_col = last_col;
          last_col = WHITE;
          break;
        case 'B':
        case 'b':
          col_balance--;
          secondlast_col = last_col;
          last_col = BLACK;
          break;
        default:
          print_error( "Invalid character in Colours", line_no );
      }
      p++;
   }
   if( secondlast_col == last_col && last_col != 0 )
   {
      if( last_col == WHITE )
         col_balance = 3;
      else
         col_balance = -3;
   }
   else
   {
      if( col_balance > 0 )
         col_balance++;
      if( col_balance < 0 )
         col_balance--;
      if( col_balance > 3 )
         col_balance = 3;
      if( col_balance < -3 )
         col_balance = -3;
   }
   if( col_balance == 0 )
   {
      if( last_col == WHITE )
         col_balance = 1;
      if( last_col == BLACK )
         col_balance = -1;
   }
   LoadColPref( player, col_balance );
   return( p );
}

//////////////////////////////////////////////////////////////////////
//
//   Extracts float status of a player
//
char *SwissRead_c::GetFloats( int player, char *p )
{
   switch( *p )
   {
      case 'U':
        LoadUpfStatus( player, LAST_FLOATED );
        break;
      case 'u':
        LoadUpfStatus( player, SECONDLAST_FLOATED );
        break;
      case 'D':
        LoadDnfStatus( player, LAST_FLOATED );
        break;
      case 'd':
        LoadDnfStatus( player, SECONDLAST_FLOATED );
        break;
      default:
        break;
   }
   return( p );
}



